@@ -79,6 +79,7 @@ enum cmd_and_opt_values
   oDebugAllowCoreDump,
   oDebugCCIDDriver,
   oDebugLogTid,
+  oDebugAllowPINLogging,
   oDebugAssuanLogCats,
   oNoGreeting,
   oNoOptions,
@@ -103,6 +104,7 @@ enum cmd_and_opt_values
   oDisableApplication,
   oApplicationPriority,
   oEnablePinpadVarlen,
+  oCompatibilityFlags,
   oListenBacklog
 };
 
@@ -138,6 +140,7 @@ static gpgrt_opt_t opts[] = {
   ARGPARSE_s_n (oDebugAllowCoreDump, "debug-allow-core-dump", "@"),
   ARGPARSE_s_n (oDebugCCIDDriver, "debug-ccid-driver", "@"),
   ARGPARSE_s_n (oDebugLogTid, "debug-log-tid", "@"),
+  ARGPARSE_s_n (oDebugAllowPINLogging, "debug-allow-pin-logging", "@"),
   ARGPARSE_p_u (oDebugAssuanLogCats, "debug-assuan-log-cats", "@"),
   ARGPARSE_s_s (oLogFile,  "log-file", N_("|FILE|write a log to FILE")),
 
@@ -170,6 +173,7 @@ static gpgrt_opt_t opts[] = {
   ARGPARSE_s_s (oDisableApplication, "disable-application", "@"),
   ARGPARSE_s_s (oApplicationPriority, "application-priority",
                 N_("|LIST|change the application priority to LIST")),
+  ARGPARSE_s_s (oCompatibilityFlags, "compatibility-flags", "@"),
   ARGPARSE_s_i (oListenBacklog, "listen-backlog", "@"),
 
 
@@ -194,6 +198,7 @@ static struct debug_flags_s debug_flags [] =
     { DBG_MEMSTAT_VALUE, "memstat" },
     { DBG_HASHING_VALUE, "hashing" },
     { DBG_IPC_VALUE    , "ipc"     },
+    { DBG_CARD_VALUE   , "card"    },
     { DBG_CARD_IO_VALUE, "cardio"  },
     { DBG_READER_VALUE , "reader"  },
     { DBG_APP_VALUE    , "app"     },
@@ -201,6 +206,14 @@ static struct debug_flags_s debug_flags [] =
   };
 
 
+/* The list of compatibility flags.  */
+static struct compatibility_flags_s compatibility_flags [] =
+  {
+    { COMPAT_CCID_NO_AUTO_DETACH, "ccid-no-auto-detach" },
+    { 0, NULL }
+  };
+
+
 /* The card driver we use by default for PC/SC.  */
 #if defined(HAVE_W32_SYSTEM) || defined(__CYGWIN__)
 #define DEFAULT_PCSC_DRIVER "winscard.dll"
@@ -236,9 +249,6 @@ static int shutdown_pending;
 /* It is possible that we are currently running under setuid permissions */
 static int maybe_setuid = 1;
 
-/* Flag telling whether we are running as a pipe server.  */
-static int pipe_server;
-
 /* Name of the communication socket */
 static char *socket_name;
 /* Name of the redirected socket or NULL.  */
@@ -271,9 +281,6 @@ static gnupg_fd_t create_server_socket (const char *name,
 static void *start_connection_thread (void *arg);
 static void handle_connections (gnupg_fd_t listen_fd);
 
-/* Pth wrapper function definitions. */
-ASSUAN_SYSTEM_NPTH_IMPL;
-
 static int active_connections;
 
 
@@ -447,6 +454,7 @@ main (int argc, char **argv )
   int greeting = 0;
   int nogreeting = 0;
   int multi_server = 0;
+  int pipe_server = 0;
   int is_daemon = 0;
   int nodetach = 0;
   int csh_style = 0;
@@ -457,7 +465,6 @@ main (int argc, char **argv )
   int allow_coredump = 0;
   struct assuan_malloc_hooks malloc_hooks;
   int res;
-  npth_t pipecon_handler;
   const char *application_priority = NULL;
 
   early_system_init ();
@@ -479,7 +486,6 @@ main (int argc, char **argv )
   malloc_hooks.free = gcry_free;
   assuan_set_malloc_hooks (&malloc_hooks);
   assuan_set_gpg_err_source (GPG_ERR_SOURCE_DEFAULT);
-  assuan_set_system_hooks (ASSUAN_SYSTEM_NPTH);
   assuan_sock_init ();
   setup_libassuan_logging (&opt.debug, NULL);
 
@@ -526,12 +532,12 @@ main (int argc, char **argv )
      Now we are working under our real uid
   */
 
-  /* The configuraton directories for use by gpgrt_argparser.  */
+  /* The configuration directories for use by gpgrt_argparser.  */
   gpgrt_set_confdir (GPGRT_CONFDIR_SYS, gnupg_sysconfdir ());
   gpgrt_set_confdir (GPGRT_CONFDIR_USER, gnupg_homedir ());
 
   /* We are re-using the struct, thus the reset flag.  We OR the
-   * flags so that the internal intialized flag won't be cleared. */
+   * flags so that the internal initialized flag won't be cleared. */
   argc = orig_argc;
   argv = orig_argv;
   pargs.argc = &argc;
@@ -586,6 +592,9 @@ main (int argc, char **argv )
         case oDebugLogTid:
           log_set_pid_suffix_cb (tid_log_callback);
           break;
+        case oDebugAllowPINLogging:
+          opt.debug_allow_pin_logging = 1;
+          break;
         case oDebugAssuanLogCats:
           set_libassuan_log_cats (pargs.r.ret_ulong);
           break;
@@ -626,6 +635,15 @@ main (int argc, char **argv )
 
         case oEnablePinpadVarlen: opt.enable_pinpad_varlen = 1; break;
 
+        case oCompatibilityFlags:
+          if (parse_compatibility_flags (pargs.r.ret_str, &opt.compat_flags,
+                                         compatibility_flags))
+            {
+              pargs.r_opt = ARGPARSE_INVALID_ARG;
+              pargs.err = ARGPARSE_PRINT_WARNING;
+            }
+          break;
+
         case oListenBacklog:
           listen_backlog = pargs.r.ret_int;
           break;
@@ -742,6 +760,7 @@ main (int argc, char **argv )
       ctrl_t ctrl;
       npth_attr_t tattr;
       gnupg_fd_t fd = GNUPG_INVALID_FD;
+      npth_t pipecon_handler;
 
 #ifndef HAVE_W32_SYSTEM
       {
@@ -757,6 +776,7 @@ main (int argc, char **argv )
       npth_init ();
       setup_signal_mask ();
       gpgrt_set_syscall_clamp (npth_unprotect, npth_protect);
+      assuan_control (ASSUAN_CONTROL_REINIT_SYSCALL_CLAMP, NULL);
 
       /* If --debug-allow-core-dump has been given we also need to
          switch the working directory to a place where we can actually
@@ -898,6 +918,7 @@ main (int argc, char **argv )
       npth_init ();
       setup_signal_mask ();
       gpgrt_set_syscall_clamp (npth_unprotect, npth_protect);
+      assuan_control (ASSUAN_CONTROL_REINIT_SYSCALL_CLAMP, NULL);
 
       /* Detach from tty and put process into a new session. */
       if (!nodetach )
@@ -997,7 +1018,7 @@ scd_deinit_default_ctrl (ctrl_t ctrl)
 /* Return the name of the socket to be used to connect to this
    process.  If no socket is available, return NULL. */
 const char *
-scd_get_socket_name ()
+scd_get_socket_name (void)
 {
   if (socket_name && *socket_name)
     return socket_name;
@@ -1005,7 +1026,36 @@ scd_get_socket_name ()
 }
 
 
-#ifndef HAVE_W32_SYSTEM
+#ifdef HAVE_W32_SYSTEM
+/* Creat an event into E_P and sets EVENTS to watch by npth_eselect.  */
+void
+scd_init_event (HANDLE *e_p, HANDLE events[2])
+{
+  HANDLE h, h2;
+  SECURITY_ATTRIBUTES sa = { sizeof (SECURITY_ATTRIBUTES), NULL, TRUE};
+
+  events[0] = *e_p = INVALID_HANDLE_VALUE;
+  events[1] = INVALID_HANDLE_VALUE;
+  /* Create event for manual reset, initially non-signaled.  Make it
+   * waitable and inheritable.  */
+  h = CreateEvent (&sa, TRUE, FALSE, NULL);
+  if (!h)
+    log_error ("can't create scd event: %s\n", w32_strerror (-1) );
+  else if (!DuplicateHandle (GetCurrentProcess(), h,
+                             GetCurrentProcess(), &h2,
+                             EVENT_MODIFY_STATE|SYNCHRONIZE, TRUE, 0))
+    {
+      log_error ("setting synchronize for scd_kick_the_loop failed: %s\n",
+                 w32_strerror (-1) );
+      CloseHandle (h);
+    }
+  else
+    {
+      CloseHandle (h);
+      events[0] = *e_p = h2;
+    }
+}
+#else
 static void
 handle_signal (int signo)
 {
@@ -1145,9 +1195,10 @@ create_server_socket (const char *name, char **r_redir_name,
     log_error (_("error getting nonce for the socket\n"));
  if (rc == -1)
     {
+      gpg_error_t myerr = gpg_error_from_syserror ();
+      log_libassuan_system_error (fd);
       log_error (_("error binding socket to '%s': %s\n"),
-                 unaddr->sun_path,
-                 gpg_strerror (gpg_error_from_syserror ()));
+                 unaddr->sun_path, gpg_strerror (myerr));
       assuan_sock_close (fd);
       scd_exit (2);
     }
@@ -1182,7 +1233,7 @@ start_connection_thread (void *arg)
       && assuan_sock_check_nonce (ctrl->thread_startup.fd, &socket_nonce))
     {
       log_info (_("error reading nonce on fd %d: %s\n"),
-                FD2INT(ctrl->thread_startup.fd), strerror (errno));
+                FD_DBG (ctrl->thread_startup.fd), strerror (errno));
       assuan_sock_close (ctrl->thread_startup.fd);
       xfree (ctrl);
       return NULL;
@@ -1193,19 +1244,20 @@ start_connection_thread (void *arg)
   scd_init_default_ctrl (ctrl);
   if (opt.verbose)
     log_info (_("handler for fd %d started\n"),
-              FD2INT(ctrl->thread_startup.fd));
+              FD_DBG (ctrl->thread_startup.fd));
 
-  /* If this is a pipe server, we request a shutdown if the command
-     handler asked for it.  With the next ticker event and given that
+  scd_command_handler (ctrl, ctrl->thread_startup.fd);
+
+  /* If this thread is the pipe connection thread, flag that a
+     shutdown is required.  With the next ticker event and given that
      no other connections are running the shutdown will then
      happen.  */
-  if (scd_command_handler (ctrl, ctrl->thread_startup.fd)
-      && pipe_server)
+  if (ctrl->thread_startup.fd == GNUPG_INVALID_FD)
     shutdown_pending = 1;
 
   if (opt.verbose)
     log_info (_("handler for fd %d terminated\n"),
-              FD2INT (ctrl->thread_startup.fd));
+              FD_DBG (ctrl->thread_startup.fd));
 
   scd_deinit_default_ctrl (ctrl);
   xfree (ctrl);
@@ -1261,7 +1313,7 @@ handle_connections (gnupg_fd_t listen_fd)
 #ifdef HAVE_PSELECT_NO_EINTR
   int pipe_fd[2];
 
-  ret = gnupg_create_pipe (pipe_fd);
+  ret = gnupg_create_pipe (pipe_fd, 0);
   if (ret)
     {
       log_error ("pipe creation failed: %s\n", gpg_strerror (ret));
@@ -1280,31 +1332,7 @@ handle_connections (gnupg_fd_t listen_fd)
   npth_attr_setdetachstate (&tattr, NPTH_CREATE_DETACHED);
 
 #ifdef HAVE_W32_SYSTEM
-  {
-    HANDLE h, h2;
-    SECURITY_ATTRIBUTES sa = { sizeof (SECURITY_ATTRIBUTES), NULL, TRUE};
-
-    events[0] = the_event = INVALID_HANDLE_VALUE;
-    events[1] = INVALID_HANDLE_VALUE;
-    /* Create event for manual reset, initially non-signaled.  Make it
-     * waitable and inheritable.  */
-    h = CreateEvent (&sa, TRUE, FALSE, NULL);
-    if (!h)
-      log_error ("can't create scd event: %s\n", w32_strerror (-1) );
-    else if (!DuplicateHandle (GetCurrentProcess(), h,
-                               GetCurrentProcess(), &h2,
-                               EVENT_MODIFY_STATE|SYNCHRONIZE, TRUE, 0))
-      {
-        log_error ("setting synchronize for scd_kick_the_loop failed: %s\n",
-                   w32_strerror (-1) );
-        CloseHandle (h);
-      }
-    else
-      {
-        CloseHandle (h);
-        events[0] = the_event = h2;
-      }
-  }
+  scd_init_event (&the_event, events);
 #endif
 
   FD_ZERO (&fdset);
@@ -1312,7 +1340,7 @@ handle_connections (gnupg_fd_t listen_fd)
   if (listen_fd != GNUPG_INVALID_FD)
     {
       FD_SET (FD2INT (listen_fd), &fdset);
-      nfd = FD2INT (listen_fd);
+      nfd = FD2NUM (listen_fd);
     }
 
   for (;;)
@@ -1398,11 +1426,13 @@ handle_connections (gnupg_fd_t listen_fd)
           gnupg_fd_t fd;
 
           plen = sizeof paddr;
-          fd = INT2FD (npth_accept (FD2INT (listen_fd),
-                                    (struct sockaddr *)&paddr, &plen));
+          fd = assuan_sock_accept (listen_fd,
+                                   (struct sockaddr *)&paddr, &plen);
           if (fd == GNUPG_INVALID_FD)
             {
-              log_error ("accept failed: %s\n", strerror (errno));
+              gpg_error_t myerr = gpg_error_from_syserror ();
+              log_libassuan_system_error (listen_fd);
+              log_error ("accept failed: %s\n", gpg_strerror (myerr));
             }
           else if ( !(ctrl = xtrycalloc (1, sizeof *ctrl)) )
             {
@@ -1416,7 +1446,7 @@ handle_connections (gnupg_fd_t listen_fd)
               npth_t thread;
 
               snprintf (threadname, sizeof threadname, "conn fd=%d",
-                        FD2INT (fd));
+                        FD_DBG (fd));
               ctrl->thread_startup.fd = fd;
               ret = npth_create (&thread, &tattr, start_connection_thread, ctrl);
               if (ret)
