@@ -33,7 +33,6 @@
 #include "../common/sysutils.h"
 #include "../common/init.h"
 #include "../common/status.h"
-#include "../common/exechelp.h"
 #include "../common/dotlock.h"
 
 #ifdef HAVE_W32_SYSTEM
@@ -434,7 +433,7 @@ valid_swdb_name_p (const char *name)
  *           'c' :: The version is Current
  *           'n' :: The current version is already Newer than the
  *                  available one.
- * urgency :: If the value is greater than zero an urgent update is required.
+ * minvers :: The minimal secure version.
  * error   :: 0 on success or an gpg_err_code_t
  *            Common codes seen:
  *            GPG_ERR_TOO_OLD :: The SWDB file is to old to be used.
@@ -465,6 +464,7 @@ query_swdb (estream_t out, const char *name, const char *current_version)
   gnupg_isotime_t filedate = {0};
   gnupg_isotime_t verified = {0};
   char *value_ver = NULL;
+  char *value_minver = NULL;
   gnupg_isotime_t value_date = {0};
   char *value_size = NULL;
   char *value_sha2 = NULL;
@@ -567,6 +567,8 @@ query_swdb (estream_t out, const char *name, const char *current_version)
             value_size = xstrdup (fields[1]);
           else if (!strcmp (p, "sha2") && !value_sha2)
             value_sha2 = xstrdup (fields[1]);
+          else if (!strcmp (p, "minver") && !value_minver)
+            value_minver = xstrdup (fields[1]);
         }
     }
   if (len < 0 || es_ferror (fp))
@@ -617,10 +619,11 @@ query_swdb (estream_t out, const char *name, const char *current_version)
   else
     status = 'n';
 
-  es_fprintf (out, "%s:%s:%c::%d:%s:%s:%s:%s:%lu:%s:\n",
+  es_fprintf (out, "%s:%s:%c:%s:%d:%s:%s:%s:%s:%lu:%s:\n",
               name,
               current_version? current_version : "",
               status,
+              value_minver? value_minver : value_ver,
               err,
               filedate,
               verified,
@@ -630,6 +633,7 @@ query_swdb (estream_t out, const char *name, const char *current_version)
               value_sha2? value_sha2 : "");
 
  leave:
+  xfree (value_minver);
   xfree (value_ver);
   xfree (value_size);
   xfree (value_sha2);
@@ -1303,7 +1307,7 @@ show_versions_via_dirmngr (estream_t fp)
   const char *pgmname;
   const char *argv[2];
   estream_t outfp;
-  gnupg_process_t proc;
+  gpgrt_process_t proc;
   char *line = NULL;
   size_t line_len = 0;
   ssize_t length;
@@ -1311,7 +1315,7 @@ show_versions_via_dirmngr (estream_t fp)
   pgmname = gnupg_module_name (GNUPG_MODULE_NAME_DIRMNGR);
   argv[0] = "--gpgconf-versions";
   argv[1] = NULL;
-  err = gnupg_process_spawn (pgmname, argv, GNUPG_PROCESS_STDOUT_PIPE,
+  err = gpgrt_process_spawn (pgmname, argv, GPGRT_PROCESS_STDOUT_PIPE,
                              NULL, &proc);
   if (err)
     {
@@ -1320,7 +1324,7 @@ show_versions_via_dirmngr (estream_t fp)
       return;
     }
 
-  gnupg_process_get_streams (proc, 0, NULL, &outfp, NULL);
+  gpgrt_process_get_streams (proc, 0, NULL, &outfp, NULL);
   while ((length = es_read_line (outfp, &line, &line_len, NULL)) > 0)
     {
       /* Strip newline and carriage return, if present.  */
@@ -1341,12 +1345,12 @@ show_versions_via_dirmngr (estream_t fp)
                  pgmname, gpg_strerror (err));
     }
 
-  err = gnupg_process_wait (proc, 1);
+  err = gpgrt_process_wait (proc, 1);
   if (!err)
     {
       int exitcode;
 
-      gnupg_process_ctl (proc, GNUPG_PROCESS_GET_EXIT_ID, &exitcode);
+      gpgrt_process_ctl (proc, GPGRT_PROCESS_GET_EXIT_ID, &exitcode);
       if (exitcode)
         {
           log_error ("running %s failed (exitcode=%d): %s\n",
@@ -1354,7 +1358,7 @@ show_versions_via_dirmngr (estream_t fp)
           es_fprintf (fp, "[error: can't get further info]\n");
         }
     }
-  gnupg_process_release (proc);
+  gpgrt_process_release (proc);
   xfree (line);
 }
 
@@ -1526,6 +1530,16 @@ show_other_registry_entries (estream_t outfp)
     { 3, "combinedOpsEnabled" },
     { 3, "encryptSubject" },
     { 0, NULL }
+    /*  We should add the following key but also hide unset ones.:
+     *   "smimeNoCertSigErr"
+     *   "smimeHtmlWarnShown"
+     *   "alwaysShowApproval"
+     *   "syncDec"
+     *   "syncEnc"
+     *   "draftEnc"
+     *   "draftKey"
+     * Or we just interate over the GpgOL keys.
+     */
   };
   int idx;
   int group = 0;
@@ -1626,7 +1640,7 @@ show_registry_entries_from_file (estream_t outfp)
       if (!any)
         {
           any = 1;
-          es_fprintf (outfp, "Taken from gpgconf.rnames:\n");
+          es_fprintf (outfp, "\nTaken from gpgconf.rnames:\n");
         }
 
       es_fprintf (outfp, "  %s\n  ->%s<-%s\n", line, value,
